#pragma once

#include <cstdint>
#include <iosfwd>
#include <stdexcept>

/* A MemoryCell forms the core of our model: all operations in our virtual
machine will manipulate MemoryCells. They are a form of "discriminated union"
(https://en.wikipedia.org/wiki/Tagged_union), which means that a MemoryCell may
hold one of several different types. It is an error to use a cell in a manner
which is inconsistent with it's actual type--for example, if I call

```
memory_cell.setPointer(ptr)
```

but memory_cell holds a Tag instead of a Pointer, the program will crash.

You can query the type of the MemoryCell with the is<TYPE>() functions. A common
programming pattern when using discriminated unions is to check the type
of the cell, then run different operations based on its type. For example:

if (cell.isTagCell()){ // operate on tag}
else if (cell.isValueCell()) { // operate on value }
else if (cell.isPointerCell()) { // operate on pointer }
else { //operate on Null cell }
*/

namespace gc {

// This enum is the type of a cell--each cell can hold one of the four types of
// data. See the project page for an explanation of how various types are used.
enum class CellType { Null, Tag, Value, Pointer };

// Only used if CellType is Tag
enum class CellTag { Int, Pair };

class RuntimeTypeError : std::exception {};

class MemoryCell {
private:
  CellType type;
  std::intptr_t data;

public:
  // Note that the constructors are overloaded by type--you will get different
  // MemoryCells depending on which constructor you call. Choose carefully!
  MemoryCell() : type(CellType::Null), data(0) {}
  MemoryCell(CellTag t)
      : type(CellType::Tag), data(static_cast<std::intptr_t>(t)) {}
  MemoryCell(std::int64_t value)
      : type(CellType::Value), data(static_cast<std::intptr_t>(value)) {}
  MemoryCell(MemoryCell *ptr)
      : type(CellType::Pointer), data(reinterpret_cast<std::intptr_t>(ptr)) {}

  CellType getCellType() const { return type; }
  bool isNullCell() const { return type == CellType::Null; }
  bool isTagCell() const { return type == CellType::Tag; }
  bool isValueCell() const { return type == CellType::Value; }
  bool isPointerCell() const { return type == CellType::Pointer; }

  // These functions are only valid if the CellType is Tag
  CellTag getTag() const;
  void setTag(CellTag);

  // These functions are only valid if the CellType is Value
  std::int64_t getValue() const;
  void setValue(std::int64_t);

  // These functions are only valid if the CellType is Pointer
  MemoryCell *getPointer() const;
  void setPointer(MemoryCell *);

  // Null out the cell
  void setNull();
};

// DO NOT EDIT THESE FUNCTIONS!!
std::ostream &operator<<(std::ostream &, CellTag);
std::ostream &operator<<(std::ostream &, CellType);
std::ostream &operator<<(std::ostream &, const MemoryCell &);
std::ostream &printGrading(std::ostream &, const MemoryCell &,
                           const MemoryCell *);
} // namespace gc